#include "../Framework2/Direct3D/Direct3DApp.h"
#include "../Framework2/Util/Direct3DModel.h"

#define DIST_SCALE 0.00028f
#define RT_SIZE 512

class MainApp : public Direct3DApp {
public:
	void initMenu();
	void resetCamera();

	bool init();
	bool exit();

	bool load();
	bool unload();

	void drawLight(const vec3 &lightPos);
	void drawDistance();
	bool drawFrame();
protected:
	Direct3DModel *model;
	vec3 lightPos;

	ShaderID lighting, lightingDXT5, lightingNoShadow, lightingDXT5NoShadow, distance;

	TextureID shadow, shadowDepth;
	TextureID light;
	TextureID base[3], bump[3], baseUncomp[3], bumpUncomp[3], bumpDXT5[3], detailBump, detailBumpUncomp, detailBumpDXT5;

	bool supports3Dc;
	bool useCompressedBase, useCompressedBump, use3Dc, drawShadows;
};

void MainApp::initMenu(){
	Menu *menu = menuSystem->getMainMenu();
	menu->addMenuItem("Use compressed base maps: ", &useCompressedBase, INPUT_BOOL);
	menu->addMenuItem("Use compressed bump maps: ", &useCompressedBump, INPUT_BOOL);
	menu->addMenuItem("Use 3Dc bump maps: ", &use3Dc, INPUT_BOOL);
	menu->addMenuItem("Draw shadows: ", &drawShadows, INPUT_BOOL);

	App::initMenu();
}

void MainApp::resetCamera(){
	position = vec3(-2000, -330, 0);
	wx = 0;
	wy = -3.1415926535f / 2;
	wz = 0;
}

bool MainApp::init(){
	model = new Direct3DModel();
	model->loadFromFile("../Models/Room2/Map.hmdl");
	for (unsigned int i = 0; i < model->getBatchCount(); i++){
		model->getBatch(i)->fixTJunctions();
	}

	useCompressedBase = true;
	useCompressedBump = true;
	use3Dc = true;
	drawShadows = true;

	return true;
}

bool MainApp::exit(){
	delete model;
	return true;
}

bool MainApp::load(){
	if (caps.PixelShaderVersion < D3DPS_VERSION(2,0) || caps.VertexShaderVersion < D3DVS_VERSION(1,1)){
		String error("This demo requires pixel shader 2.0 and vertex shader 1.1.\n"
					 "Your card/driver only supports pixel shader ");
		error.sprintf("%d.%d", (caps.PixelShaderVersion  >> 8) & 255, caps.PixelShaderVersion  & 255);
		error += " and vertex shader ";
		error.sprintf("%d.%d", (caps.VertexShaderVersion >> 8) & 255, caps.VertexShaderVersion & 255);

		addToLog(error);
		return false;
	}

	supports3Dc = (d3d->CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8, 0, D3DRTYPE_TEXTURE, (D3DFORMAT) '2ITA') == D3D_OK);

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	//shadow = renderer->addRenderCubemap(RT_SIZE, FORMAT_RG16);
	shadow = renderer->addRenderCubemap(RT_SIZE, FORMAT_I16F);
	shadowDepth = renderer->addRenderDepth(RT_SIZE, RT_SIZE, FORMAT_D16);

	if ((lighting = renderer->addShader("lighting.shd")) == SHADER_NONE) return false;
	if ((lightingDXT5 = renderer->addShader("lighting.shd", "#define USE_DXT5\n")) == SHADER_NONE) return false;
	if ((lightingNoShadow = renderer->addShader("lighting.shd", "#define NO_SHADOWS")) == SHADER_NONE) return false;
	if ((lightingDXT5NoShadow = renderer->addShader("lighting.shd", "#define USE_DXT5\n#define NO_SHADOWS\n")) == SHADER_NONE) return false;

	if ((distance = renderer->addShader("distance.shd")) == SHADER_NONE) return false;

	char *baseMaps[] = {
		"../Textures/board4DXT5.dds",
		"../Textures/parqfloor2DXT5.dds",
		"../Textures/laying_rock7_largeDXT5.dds"
	};

	char *bumpMaps[] = {
		"../Textures/board4Bump3Dc.dds",
		"../Textures/parqfloor2Bump3Dc.dds",
		"../Textures/laying_rock7Bump3Dc.dds"
	};

	char *bumpMapsDXT5[] = {
		"../Textures/board4BumpDXT5.dds",
		"../Textures/parqfloor2BumpDXT5.dds",
		"../Textures/laying_rock7BumpDXT5.dds"
	};

	char *detailBumpMap     = "../Textures/genericDetailBump3Dc.dds";
	char *detailBumpMapDXT5 = "../Textures/genericDetailBumpDXT5.dds";

	Image *img = new Image();
	img->loadImage(detailBumpMap, true, MIPMAP_CREATE);
	img->uncompressImage(false);
	if ((detailBumpUncomp = renderer->addTexture(img)) == TEXTURE_NONE) return false;

	if (supports3Dc){
		if ((detailBump = renderer->addTexture(detailBumpMap)) == TEXTURE_NONE) return false;
	}
	if ((detailBumpDXT5 = renderer->addTexture(detailBumpMapDXT5)) == TEXTURE_NONE) return false;

	for (int i = 0; i < 3; i++){
		if ((base[i] = renderer->addTexture(baseMaps[i])) == TEXTURE_NONE) return false;
		if (supports3Dc){
			if ((bump[i] = renderer->addTexture(bumpMaps[i])) == TEXTURE_NONE) return false;
		}
		if ((bumpDXT5[i] = renderer->addTexture(bumpMapsDXT5[i])) == TEXTURE_NONE) return false;

		img = new Image();
		img->loadImage(baseMaps[i], true, MIPMAP_CREATE);
		img->uncompressImage(false);
		if ((baseUncomp[i] = renderer->addTexture(img)) == TEXTURE_NONE) return false;
	
		img = new Image();
		img->loadImage(bumpMaps[i], true, MIPMAP_CREATE);
		img->uncompressImage(false);
		if ((bumpUncomp[i] = renderer->addTexture(img)) == TEXTURE_NONE) return false;
	}

	if ((light = renderer->addTexture("../Textures/Particle.tga")) == TEXTURE_NONE) return false;

	model->uploadToVertexBuffer(dev);

	return true;
}

bool MainApp::unload(){
	model->freeVertexBuffer();
	return true;
}

void MainApp::drawLight(const vec3 &lightPos){
	renderer->setTextures(light);
	renderer->setBlending(ONE, ONE);
	renderer->apply();

	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);

	struct Vertex {
		vec3 position;
		vec2 texCoord;
	};

	Vertex vertices[4];

	vertices[0].position = lightPos + 30 * (-dx + dy);
	vertices[1].position = lightPos + 30 * ( dx + dy);
	vertices[2].position = lightPos + 30 * ( dx - dy);
	vertices[3].position = lightPos + 30 * (-dx - dy);

	vertices[0].texCoord = vec2(0, 0);
	vertices[1].texCoord = vec2(1, 0);
	vertices[2].texCoord = vec2(1, 1);
	vertices[3].texCoord = vec2(0, 1);


	dev->SetFVF(D3DFVF_XYZ | D3DFVF_TEXCOORDSIZE2(0) | (1 << D3DFVF_TEXCOUNT_SHIFT));
	dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, sizeof(Vertex));
}

void MainApp::drawDistance(){
	renderer->setShader(distance);
	renderer->apply();

	renderer->changeShaderConstant3f("lightPos", lightPos);
	renderer->changeShaderConstant1f("scale", DIST_SCALE);

	mat4 proj = projectionMatrixX(PI / 2, 1, 1, 3000);

	for (int face = 0; face < 6; face++){
		renderer->changeRenderTarget(shadow, shadowDepth, face);
		dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(255, 255, 255, 255), 1.0f, 0);

		mat4 view = cubemapMatrix(face);
		view.translate(-lightPos);
		renderer->changeShaderConstant4x4f("view_proj_matrix", proj * view);

		((Direct3DModel *) model)->draw(dev);
	}
	renderer->changeToDefaultRenderTarget();
}

bool MainApp::drawFrame(){
	if (!supports3Dc){
		if (use3Dc){
			Menu *menu = menuSystem->getMainMenu();
			menu->nextValue(2);
		}
	}

    dev->Clear(0, NULL, /*D3DCLEAR_TARGET | */D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0, 0, 0, 0), 1.0f, 0);

	projection = projectionMatrixX(1.5f, float(height) / float(width), 1, 6000);
	dev->SetTransform(D3DTS_PROJECTION, (D3DMATRIX *) (const float *) transpose(projection));

	modelView = rotateZXY(-wx, -wy, -wz);
	modelView.translate(-position);
	dev->SetTransform(D3DTS_VIEW, (D3DMATRIX *) (const float *) transpose(modelView));

	dev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	lightPos = vec3(800 * cosf(time), 650 * cosf(0.7f * time), 300 * sinf(time));

	if (drawShadows) drawDistance();

	float pSize[] = { 0.01f, 0.02f, 0.03f };
	float texWorldSize[] = { 1024, 256, 1024 };
	for (unsigned int i = 0; i < model->getBatchCount(); i++){
		if (use3Dc || !useCompressedBump){
			renderer->setShader(drawShadows? lighting : lightingNoShadow);
		} else {
			renderer->setShader(drawShadows? lightingDXT5 : lightingDXT5NoShadow);
		}

		renderer->setTexture("DetailBump", useCompressedBump? (use3Dc? detailBump : detailBumpDXT5) : detailBumpUncomp);
		renderer->setTexture("Bump", useCompressedBump? (use3Dc? bump[i] : bumpDXT5[i]) : bumpUncomp[i]);
		renderer->setTexture("Base", useCompressedBase? base[i] : baseUncomp[i]);
		if (drawShadows) renderer->setTexture("Shadow", shadow);

		renderer->apply();

		renderer->changeShaderConstant4x4f("view_proj_matrix", projection * modelView);
		renderer->changeShaderConstant3f("lightPos", lightPos);
		renderer->changeShaderConstant3f("camPos", position);
		renderer->changeShaderConstant2f("pSize", vec2(pSize[i] * 2, pSize[i]));
		renderer->changeShaderConstant1f("scale", DIST_SCALE);
		renderer->changeShaderConstant1f("texWorldSize", texWorldSize[i]);

		((Direct3DBatch *) model->getBatch(i))->draw(dev);
	}

	drawLight(lightPos);

	return true;
}

App *app = new MainApp();
